#include <ctype.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>

#include "list.h"

struct ctrie_node
   {
   char * key;
   struct ctrie_node * parent;
   struct ctrie_node * child;
   struct ctrie_node * next_sibling;
   struct ctrie_node * prev_sibling;
   void * data;
   };

struct ctrie
   {
   struct ctrie_node * root;
   size_t size;
   size_t count;
   };


static struct list * subkey_list = NULL;
static size_t subkey_list_limit = 0;

char *              ctrie_insert_node(struct ctrie * ctrie, struct ctrie_node * node, char * key, void * data);
struct ctrie_node * ctrie_longest_matching_child(struct ctrie_node * node, char * key);
struct ctrie_node * ctrie_node_new(char * key, void * data);
int                 ctrie_str_common(char * s1, char * s2);

void ctrie_print_node(struct ctrie_node * node);
void ctrie_print_key (struct ctrie_node * node);
void ctrie_node_dump (struct ctrie_node * node, int depth, void (*op)(struct ctrie_node *));

void ctrie_visit_node(struct ctrie * ctrie, struct ctrie_node * node, void (*op)(char * k, void * d), char * key);

struct ctrie_node * ctrie_find_node      (struct ctrie_node * node, char * key);
struct ctrie_node * ctrie_search_children(struct ctrie_node * node, char * k);


void ctrie_subtree_list_push(char * k, void * d);

////////////////////////////////////////////////////////////////////////////////
struct ctrie *
ctrie_new()
   {
   struct ctrie * ctrie = malloc(sizeof(struct ctrie));
   if (NULL == ctrie)
      return NULL;

   ctrie->root = malloc(sizeof(struct ctrie_node));
   if (NULL == ctrie->root)
      return NULL;

   ctrie->root->parent = ctrie->root->child = ctrie->root->next_sibling = NULL;
   ctrie->root->key = "";
   ctrie->root->data = NULL;
   ctrie->size = ctrie->count = 0;

   printf("The root node is %p\n", ctrie->root);
   return ctrie;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_delete(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return;

   free(ctrie);
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_node_new(char * key, void * data)
   {
   if (NULL == key)
      return NULL;

   struct ctrie_node * new = malloc(sizeof(struct ctrie_node));
   if (NULL == new)
      return NULL;

   new->key = malloc(strlen(key));
   if (NULL == new->key)
      {
      free(new);
      return NULL;
      }
   strcpy(new->key, key);

   new->data = data;
   new->parent = new->child = new->prev_sibling = new->next_sibling = NULL;

   return new;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ctrie_node_free(struct ctrie_node * node)
   {
   if (NULL == node)
      return NULL;
   void * data = node->data;
   free(node->key);
   free(node);
   return data;
   }

////////////////////////////////////////////////////////////////////////////////
size_t
ctrie_size(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return 0;
   return ctrie->size;
   }

////////////////////////////////////////////////////////////////////////////////
size_t
ctrie_count(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return 0;
   return ctrie->count;
   }

////////////////////////////////////////////////////////////////////////////////
char *
ctrie_insert(struct ctrie * ctrie, char * key, void * data)
   {
   if ( (NULL == ctrie) || (NULL == key) )
      return NULL;

   printf("root node is ");
   ctrie_print_node(ctrie->root);
   return ctrie_insert_node(ctrie, ctrie->root, key, data);
   }

////////////////////////////////////////////////////////////////////////////////
char *
ctrie_insert_node(struct ctrie * ctrie, struct ctrie_node * node, char * key, void * data)
   {
   if ( (NULL == node) || (NULL == key) )
      return NULL;

   struct ctrie_node * p = ctrie_longest_matching_child(node, key);

   /*
   printf("Longest match is ");
   if (NULL != p)
      ctrie_print_node(p);
   else
      printf("(nil)\n");
   */

   if (NULL == p)
      {
      // No matching child found.  Create new(key, data) as child of node
      struct ctrie_node * new = ctrie_node_new(key, data);
      if (NULL == new)
	 return NULL;
      new->parent = node;

      if (NULL == node->child)
	 node->child = new;
      else 
	 {
	 struct ctrie_node * s = node->child;
	 while (NULL != s->next_sibling)
	    s = s->next_sibling;
	 s->next_sibling = new;
	 new->prev_sibling = s;
	 }
      
      ctrie->size++;
      ctrie->count++;

      /*
      printf("Added ");
      ctrie_print_node(new);
      */

      return key;
      }
   else if (ctrie_str_common(p->key, key) == strlen(p->key))
      {
      // Found a node whose key matches for it's entire length
      if (0 == strcmp(p->key, key))
	 {
	 // Key already exists.  Error
	 return NULL;
	 }
      else
	 {
	 // The node found has a key that is a prefix of the key we are inserting.
	 // Recurse with the remainder of the key
	 //	 printf("Recursing because node's entire key %s is a prefix of %s\n", p->key, key);
	 return ctrie_insert_node(ctrie, p, key+strlen(p->key), data);
	 }
      }
   else
      {
      // Found a node that matches for only part of it's length.
      // Split the node and then recurse.

      /*
      printf("splitting this node: ");
      ctrie_print_node(p);
      */

      // Get the new shorter key for p.
      int matchlen = ctrie_str_common(p->key, key);
      char * prefix = malloc(matchlen+1);
      if (NULL == prefix)
	 return NULL;

      strncpy(prefix, key, matchlen);
      prefix[matchlen] = '\0';

      // Create a new node to hold the remainder of the old key and take over the existing children.
      struct ctrie_node * new = ctrie_node_new((p->key)+matchlen, p->data);
      if (NULL == new)
	 {
	 free(prefix);
	 return NULL;
	 }
      new->parent = p;
      new->child = p->child;
      new->data = p->data;

      // Update p to point to it's new child and have it's new (shorter) key
      free(p->key);
      p->key = prefix;
      p->child = new;
      p->data = NULL;

      ctrie->size++;

      /*
      printf("Node was split into these\n");
      ctrie_print_node(p);
      ctrie_print_node(new);
      */

      // Now add the user requested new key as a child of p
      //      printf("Recursing with key %s as a prefix of %s\n", p->key, key);
      return ctrie_insert_node(ctrie, p, key + matchlen, data);
      }

   // should not get here
   return NULL;
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_longest_matching_child(struct ctrie_node * node, char * key)
   {
   struct ctrie_node * longest = NULL;
   int len = 0;
   int l;

   printf("Searching for match in children of %p\n", node);

   struct ctrie_node * p = node->child;
   while (NULL != p)
      {
      l = ctrie_str_common(p->key, key);
      printf("common length is %d\n", l);
      if (l > len)
	 {
	 len = l;
	 longest = p;
	 }
      p = p->next_sibling;
      }
   return longest;
   }

////////////////////////////////////////////////////////////////////////////////
// Returns the number of characters of s1 (starting at the beginning)
// that match the corresponding characters in s2.
//
// eg: ctrie_str_common("asdf", "asdf") returns 4
//     ctrie_str_common("asdf", "asdfer") returns 4
//     ctrie_str_common("asdf", "asgg") returns 2
//     ctrie_str_common("asdf", "gggg") returns 0

int
ctrie_str_common(char * s1, char * s2)
   {
   int i = 0;
   while ( (*s1 != '\0') && (*s2 != '\0') &&
	   (*(s1+i) == *(s2+i)) )
      i++;

   return i;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_print_node(struct ctrie_node * node)
   {
   if (NULL == node)
      printf("(nil node)\n");
   else
      printf("%s \t%p\t parent:%p \tchild:%p \tprev:%p \tnext:%p \tdata:%p\n", node->key, node, node->parent, node->child, node->prev_sibling, node->next_sibling, node->data);

   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_print_key(struct ctrie_node * node)
   {
   if (NULL == node)
      printf("(nil node)\n");
   else if (NULL == node->key)
      printf("(nil key)\n");
   else
      printf("%s\n", node->key);
   }

////////////////////////////////////////////////////////////////////////////////
void ctrie_dump(struct ctrie * ctrie)
   { 
   if (NULL == ctrie)
      return;
  
   int depth = 0;
   ctrie_node_dump(ctrie->root, depth, ctrie_print_key);
   }

////////////////////////////////////////////////////////////////////////////////
void ctrie_dump_raw(struct ctrie * ctrie)
   { 
   if (NULL == ctrie)
      return;
  
   int depth = 0;
   ctrie_node_dump(ctrie->root, depth, ctrie_print_node);
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_node_dump(struct ctrie_node * node, int depth, void (*op)(struct ctrie_node *))
   {
   for (int i=0; i < depth; i++)
      printf(" ");
   op(node);
   if (node->child)
      {
      int len = 0;
      if (NULL != node->key)
	 len = strlen(node->key);
      ctrie_node_dump(node->child, depth+len, op); 
      }
   if (NULL != node->next_sibling)
      ctrie_node_dump(node->next_sibling, depth, op);
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_walk_keys(struct ctrie * ctrie, void (*op)(char * k, void * d))
   {
   if (NULL == ctrie)
      return;
   ctrie_visit_node(ctrie, ctrie->root, op, "");
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_visit_node(struct ctrie * ctrie, struct ctrie_node * node, void (*op)(char * k, void * d), char * prefix)
   {
   if (NULL != node->data)
      {
      char * k = malloc(strlen(prefix) + strlen(node->key));
      if (NULL == k)
	 return;
      k[0] = '\0';
      strcat(k, prefix);
      strcat(k, node->key);
      op(k, node->data);
      free(k);
      }

   if(NULL != node->child)
      {
      char * k = malloc(strlen(prefix) + strlen(node->key));
      if (NULL == k)
	 return;
      k[0] = '\0';
      strcat(k, prefix);
      strcat(k, node->key);
      struct ctrie_node * p = node->child;
      while (NULL != p)
	 {
	 ctrie_visit_node(ctrie, p, op, k);
	 p = p->next_sibling;
	 }
      free(k);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void *
ctrie_find(struct ctrie * ctrie, char * key)
   {
   struct ctrie_node * node = ctrie_find_node(ctrie->root, key);
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_find_node(struct ctrie_node * node, char * key)
   {
   if ( (NULL == node) || (NULL == key) )
      return NULL;

   size_t len = strlen(key);
   for (int i = 0; i <= len; i += strlen(node->key))
      {
      node = ctrie_search_children(node, key+i);
      if (NULL == node)
	 return NULL;
      if (NULL != node->data)
	 return node;
      }

   // Should never get here
   return NULL;

   struct ctrie_node * p = ctrie_longest_matching_child(node, key);
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_search_children(struct ctrie_node * node, char * k)
   {
   struct ctrie_node * p = node->child;
   if (NULL == p)
      return NULL;

   size_t len = strlen(node->key);
   while ( (NULL != p) && (0 != strncmp(node->key, k, len)) )
      p = p->next_sibling;
   return p;
   }




















////////////////////////////////////////////////////////////////////////////////
void *
ctrie_remove(struct ctrie * ctrie, char * key)
   {
   if ( (NULL == ctrie) || (NULL == key) )
      return NULL;

   struct ctrie_node * node = ctrie_find_node(ctrie, key);
   if (NULL == node)
      return NULL;

   void * data = node->data;

   while ( (NULL == node->prev_sibling) && (NULL == node->prev_sibling) ) 
      {
      struct ctrie_node * p = node;
      node = node->parent;
      free(p);
      ctrie->size--;
      }

   if (node == ctrie->root)
      return data;

   if (node->prev_sibling)
      node->prev_sibling->next_sibling = node->next_sibling;
   else
      node->parent->child = node->next_sibling;

   if (node->next_sibling)
      node->next_sibling->prev_sibling = node->prev_sibling;

   free(node);

   ctrie->count--;

   return data;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_subtree_list_push(char * k, void * d)
   {
   if (NULL == subkey_list)
      return;

   if ( (0 == subkey_list_limit) || (list_size(subkey_list) < subkey_list_limit) )
      list_push_front(subkey_list, k);
   }

////////////////////////////////////////////////////////////////////////////////
// Not reentrant since it uses the package static subkeys_list variable.
// However, that variable is only used while building a subkey list, so you can
// get more than one subkey if you want.  Just don't try to get them simultaneously.
struct list *
ctrie_get_subkeys(struct ctrie * ctrie, char * prefix, size_t limit)
   {
   if ( (NULL == ctrie) || (NULL == prefix) )
      return NULL;

   struct ctrie_node * node = ctrie_find_node(ctrie, prefix);
   if (NULL == node)
      return NULL;

   subkey_list = list_new();
   if (NULL == subkey_list)
      return NULL;
   subkey_list_limit = limit;

   size_t len = strlen(prefix);
   char * pref = malloc(len+1);
   if (NULL == pref)
      return NULL;
   strncpy(pref, prefix, len-1);
   pref[len-1] = '\0';

   ctrie_visit_node(ctrie, node->parent, ctrie_subtree_list_push, pref);
   free(pref);

   return subkey_list;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_free_subkey_list(struct list * list)
   {
   if (NULL == list)
      return;

   struct list_iterator * it = list_begin(list);
   void * data;
   while (data = list_remove(it))
      free(data);

   list_delete(list);
   }
